"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var middleware_client_1 = require("../service/middleware-client");
var datastore_service_1 = require("../service/datastore-service");
var methodCleaner_1 = require("../service/data-processor/methodCleaner");
var cleaner_1 = require("../service/data-processor/cleaner");
var diff_1 = require("../service/data-processor/diff");
var null_1 = require("../service/data-processor/null");
var _ = require("lodash");
var model_descriptor_service_1 = require("../service/model-descriptor-service");
var immutable_1 = require("immutable");
var AbstractDao = (function () {
    function AbstractDao(objectType, config) {
        this.isRegistered = false;
        config = config || {};
        this.objectType = config.typeName || objectType;
        this.middlewareName = config.middlewareName || _.kebabCase(objectType);
        this.queryMethod = config.queryMethod || AbstractDao.dotCase(objectType) + '.query';
        this.updateMethod = config.updateMethod || AbstractDao.dotCase(objectType) + '.update';
        this.createMethod = config.createMethod || AbstractDao.dotCase(objectType) + '.create';
        this.deleteMethod = config.deleteMethod || AbstractDao.dotCase(objectType) + '.delete';
        this.eventName = config.eventName || 'entity-subscriber.' + this.middlewareName + '.changed';
        this.idPath = config.idPath || 'id';
        this.preventQueryCaching = config.preventQueryCaching;
        this.middlewareClient = middleware_client_1.MiddlewareClient.getInstance();
        this.datastoreService = datastore_service_1.DatastoreService.getInstance();
        this.modelDescriptorService = model_descriptor_service_1.ModelDescriptorService.getInstance();
        this.taskDescriptorsPromise = new Map();
    }
    AbstractDao.prototype.list = function (partial, criteria) {
        var _this = this;
        if (partial === void 0) { partial = false; }
        return (this.listPromise && !this.preventQueryCaching) ?
            this.listPromise :
            this.listPromise = this.stream(partial, criteria).then(function (stream) {
                _this.register();
                var dataArray = stream.get('data').toJS();
                dataArray._objectType = _this.objectType;
                return dataArray;
            });
    };
    AbstractDao.prototype.stream = function (partial, criteria) {
        if (partial === void 0) { partial = false; }
        var middlewareCriteria = criteria ? this.getMiddlewareCriteria(criteria) : [];
        return this.datastoreService.stream(this.objectType, this.queryMethod, this.idPath, partial, middlewareCriteria);
    };
    AbstractDao.prototype.getNextSequenceForStream = function (streamId) {
        return this.datastoreService.getNextSequenceForStream(streamId);
    };
    AbstractDao.prototype.get = function () {
        var _this = this;
        return this.getPromise ? this.getPromise : this.query().then(function (x) {
            _this.getPromise = null;
            return x[0];
        });
    };
    AbstractDao.prototype.findSingleEntry = function (criteria, params) {
        params = params || {};
        params.single = true;
        return this.query(criteria, params).then(function (results) {
            return _.assign(results[0], { _stableId: results[0].id });
        });
    };
    AbstractDao.prototype.find = function (criteria) {
        criteria = criteria || {};
        return this.stream(false, criteria).then(function (results) {
            return results.get('data');
        });
    };
    AbstractDao.prototype.register = function () {
        if (!this.isRegistered) {
            this.middlewareClient.subscribeToEvents(this.eventName);
            this.isRegistered = true;
        }
    };
    AbstractDao.prototype.save = function (object, args) {
        return object._isNew ? this.create(object, args) : this.update(object, args);
    };
    AbstractDao.prototype.delete = function (object, args) {
        args = args || [];
        return this.middlewareClient.submitTask(this.deleteMethod, _.concat([object.id], args));
    };
    AbstractDao.prototype.getNewInstance = function () {
        var self = this;
        return Promise.resolve(new Object({
            _isNew: true,
            _objectType: self.objectType
        }));
    };
    AbstractDao.prototype.getEmptyList = function () {
        var self = this;
        var emptyList = [];
        emptyList._objectType = self.objectType;
        return Promise.resolve(emptyList);
    };
    AbstractDao.prototype.revert = function (object) {
        var reference = this.datastoreService.getState().get(object._objectType).get(object.id);
        _.forEach(object, function (value, key) {
            if (key[0] !== '_') {
                if (reference.has(key)) {
                    var referenceValue = reference.get(key);
                    if (referenceValue instanceof immutable_1.Map || referenceValue instanceof immutable_1.List) {
                        referenceValue = referenceValue.toJS();
                    }
                    if (value !== referenceValue) {
                        object[key] = referenceValue;
                    }
                }
            }
        });
        return object;
    };
    AbstractDao.prototype.update = function (object, args) {
        var _this = this;
        args = args || [];
        return Promise.all([
            this.loadPropertyDescriptors(),
            this.loadTaskDescriptor(this.updateMethod)
        ]).spread(function (propertyDescriptors, methodDescriptor) {
            var update = methodCleaner_1.processor.process(diff_1.processor.process(cleaner_1.processor.process(object, propertyDescriptors), _this.objectType, object.id, propertyDescriptors), methodDescriptor);
            if (update || (args && args.length > 0)) {
                var payload = _.concat(object._stableId ? [object._stableId, update] : [update], args);
                return _this.middlewareClient.submitTask(_this.updateMethod, payload).then(function (submittedTask) {
                    _this.datastoreService.saveOverlay(_this.objectType, object._stableId, submittedTask.taskId, _.assign(update, { _isLocked: true }));
                    submittedTask.taskPromise.then(function () {
                        _this.datastoreService.deleteOverlay(_this.objectType, object._stableId, submittedTask.taskId);
                    });
                    return submittedTask;
                });
            }
        });
    };
    AbstractDao.prototype.create = function (object, args) {
        var _this = this;
        args = args || [];
        return Promise.all([
            this.loadPropertyDescriptors(),
            this.loadTaskDescriptor(this.createMethod)
        ]).spread(function (propertyDescriptors, methodDescriptor) {
            var newObject = methodCleaner_1.processor.process(null_1.processor.process(cleaner_1.processor.process(object, propertyDescriptors)), methodDescriptor);
            if (newObject) {
                return _this.middlewareClient.submitTask(_this.createMethod, _.concat([newObject], args)).then(function (submittedTask) {
                    _this.datastoreService.saveOverlay(_this.objectType, object._stableId, submittedTask.taskId, newObject);
                    return submittedTask;
                });
            }
        });
    };
    AbstractDao.prototype.query = function (criteria, isSingle) {
        var self = this, middlewareCriteria = criteria ? this.getMiddlewareCriteria(criteria, isSingle) : [];
        return this.datastoreService.query(self.objectType, self.queryMethod, this.idPath, middlewareCriteria).then(function (entries) {
            entries = Array.isArray(entries) ? entries : [entries];
            self.register();
            var results = entries.map(function (x) {
                x._objectType = self.objectType;
                return x;
            });
            results._objectType = self.objectType;
            return results;
        });
    };
    AbstractDao.prototype.getMiddlewareCriteria = function (criteria, params) {
        var _this = this;
        var keys = Object.keys(criteria), middlewareCriteria = [], key, values;
        for (var i = 0, length_1 = keys.length; i < length_1; i++) {
            key = keys[i];
            values = _.castArray(criteria[key]);
            _.forEach(values, function (value) {
                if (typeof value === 'object') {
                    if (Array.isArray(value)) {
                        middlewareCriteria.push(_.flatten([key, value]));
                    }
                    else {
                        var subCriteria = _.flatten(_this.getMiddlewareCriteria(value));
                        Array.prototype.push.apply(middlewareCriteria, subCriteria.map(function (x) { return [key + '.' + x[0], x[1], x[2]]; }));
                    }
                }
                else {
                    middlewareCriteria.push([key, '=', value]);
                }
            });
        }
        return params ? [middlewareCriteria, params] : [middlewareCriteria];
    };
    AbstractDao.dotCase = function (aString) {
        return _.replace(_.snakeCase(aString), '_', '.');
    };
    AbstractDao.prototype.loadTaskDescriptor = function (method) {
        if (!this.taskDescriptorsPromise.has(method)) {
            this.taskDescriptorsPromise.set(method, this.modelDescriptorService.getTaskDescriptor(method));
        }
        return this.taskDescriptorsPromise.get(method);
    };
    AbstractDao.prototype.loadPropertyDescriptors = function () {
        if (!this.propertyDescriptorsPromise) {
            this.propertyDescriptorsPromise = this.modelDescriptorService.getPropertyDescriptorsForType(this.objectType);
        }
        return this.propertyDescriptorsPromise;
    };
    return AbstractDao;
}());
exports.AbstractDao = AbstractDao;
