"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var crossroads = require("crossroads");
var event_dispatcher_service_1 = require("../service/event-dispatcher-service");
var model_descriptor_service_1 = require("../service/model-descriptor-service");
var model_event_name_1 = require("../model-event-name");
var data_object_change_service_1 = require("../service/data-object-change-service");
var model_1 = require("../model");
var AbstractRoute = (function () {
    function AbstractRoute(eventDispatcherService, modelDescriptorService, dataObjectChangeService) {
        if (eventDispatcherService === void 0) { eventDispatcherService = event_dispatcher_service_1.EventDispatcherService.getInstance(); }
        if (modelDescriptorService === void 0) { modelDescriptorService = model_descriptor_service_1.ModelDescriptorService.getInstance(); }
        if (dataObjectChangeService === void 0) { dataObjectChangeService = new data_object_change_service_1.DataObjectChangeService(); }
        this.eventDispatcherService = eventDispatcherService;
        this.modelDescriptorService = modelDescriptorService;
        this.dataObjectChangeService = dataObjectChangeService;
        this.isFirstTime = true;
        this.isCurrentSection = false;
    }
    AbstractRoute.prototype.updateStackWithContext = function (stack, context) {
        this.popStackAtIndex(stack, context.columnIndex);
        stack.push(context);
        return stack;
    };
    AbstractRoute.prototype.loadObjectInColumn = function (stack, columnIndex, previousColumnIndex, pathSuffix, objectType, dataPromise, filter, postProcessor) {
        var _this = this;
        if (postProcessor === void 0) { postProcessor = _.noop; }
        var parentContext = stack[previousColumnIndex], context = {
            columnIndex: columnIndex,
            objectType: objectType,
            parentContext: parentContext,
            path: parentContext.path + pathSuffix
        }, promise = Promise.all([
            dataPromise,
            this.modelDescriptorService.getUiDescriptorForType(objectType)
        ]).spread(function (objects, uiDescriptor) {
            context.object = filter ? _.find(objects, filter) : objects;
            context.userInterfaceDescriptor = uiDescriptor;
            context.objectType = objectType;
            return context;
        }).disposer(function () {
            _this.updateStackWithContext(stack, context);
        });
        Promise.using(promise, postProcessor);
        return promise;
    };
    AbstractRoute.prototype.loadPropertyInColumn = function (stack, columnIndex, previousColumnIndex, pathSuffix, objectType, propertyPath) {
        var _this = this;
        var parentContext = stack[previousColumnIndex], context = {
            columnIndex: columnIndex,
            objectType: objectType,
            parentContext: parentContext,
            path: parentContext.path + pathSuffix
        };
        return Promise.all([
            this.modelDescriptorService.getUiDescriptorForType(objectType)
        ]).spread(function (uiDescriptor) {
            context.object = _.get(parentContext.object, propertyPath);
            context.userInterfaceDescriptor = uiDescriptor;
            context.objectType = objectType;
            return _this.updateStackWithContext(stack, context);
        });
    };
    AbstractRoute.prototype.loadListInColumn = function (stack, columnIndex, previousColumnIndex, pathSuffix, objectType, dataPromise, options) {
        var _this = this;
        options = options || {};
        var parentContext = stack[previousColumnIndex], context = {
            columnIndex: columnIndex,
            objectType: objectType,
            parentContext: parentContext,
            path: parentContext.path + pathSuffix
        };
        return Promise.all([
            dataPromise,
            this.modelDescriptorService.getUiDescriptorForType(objectType)
        ]).spread(function (objects, uiDescriptor) {
            var filteredObjects = _.sortBy(_.filter(objects, options.filter || _.identity), options.sort || ['name', 'id']);
            context.objectType = filteredObjects._objectType = objectType;
            context.object = filteredObjects;
            context.userInterfaceDescriptor = uiDescriptor;
            context.changeListener = _this.eventDispatcherService.addEventListener(model_event_name_1.ModelEventName[objectType].listChange, function (state) {
                return _this.dataObjectChangeService.handleDataChange(filteredObjects, state, options);
            });
            return _this.updateStackWithContext(stack, context);
        });
    };
    AbstractRoute.getObjectPathSuffix = function (model, id) {
        return '/' + _.kebabCase(model) + '/_/' + encodeURIComponent(_.toString(id));
    };
    AbstractRoute.prototype.popStackAtIndex = function (stack, index) {
        while (stack.length > index) {
            var context = stack.pop();
            if (context) {
                this.unregisterChangeListeners(context.changeListener);
            }
        }
    };
    AbstractRoute.prototype.unregisterChangeListeners = function (changeListeners) {
        if (changeListeners) {
            for (var _i = 0, _a = _.castArray(changeListeners); _i < _a.length; _i++) {
                var listener = _a[_i];
                this.eventDispatcherService.removeEventListener(listener.eventName, listener);
            }
        }
    };
    AbstractRoute.prototype.restore = function (stack) {
        this.stack = stack;
    };
    AbstractRoute.prototype.getStack = function () {
        return this.stack;
    };
    AbstractRoute.prototype.saveState = function () {
        var stateSnapshot = [];
        _.forEach(this.stack, function (value, index) {
            var context = _.clone(value);
            if (index > 0) {
                context.parentcontext = stateSnapshot[index - 1];
            }
            stateSnapshot.push(context);
        });
        return stateSnapshot;
    };
    AbstractRoute.prototype.enterSection = function (sectionId) {
        var _this = this;
        if (this.isFirstTime) {
            this.isFirstTime = false;
            crossroads.routed.add(function (request) {
                _this.isCurrentSection = AbstractRoute.getPathSection(request) === sectionId;
            });
        }
        var promise = (this.stack && !this.isCurrentSection) ?
            new Promise(function (resolve) { return resolve(_this.stack); }) :
            this.getStackForSection(sectionId).then(function (stack) { return _this.stack = stack; });
        promise.then(function (stack) {
            _this.eventDispatcherService.dispatch('sectionChange', stack[0].service);
            _this.eventDispatcherService.dispatch('pathChange', stack);
        });
    };
    AbstractRoute.prototype.getStackForSection = function (sectionId) {
        var self = this, objectType = model_1.Model.Section, sectionDescriptor;
        return this.loadSectionsDescriptors().then(function (sectionsDescriptors) {
            sectionDescriptor = sectionsDescriptors[sectionId];
            return Promise.resolve(AbstractRoute.sectionsServices.has(sectionDescriptor.id) ?
                AbstractRoute.sectionsServices.get(sectionDescriptor.id) :
                require.async(sectionDescriptor.service).then(function (module) {
                    var exports = Object.keys(module);
                    if (exports.length === 1) {
                        var clazz = module[exports[0]], instance = clazz.instance || new clazz(), instancePromise = instance.instanciationPromise;
                        AbstractRoute.sectionsServices.set(sectionDescriptor.id, instance);
                        return instancePromise;
                    }
                }).then(function (service) {
                    service.sectionGeneration = 'new';
                    service.section.id = sectionDescriptor.id;
                    service.section.settings.id = sectionDescriptor.id;
                    service.section.label = sectionDescriptor.label;
                    service.section.icon = sectionDescriptor.icon;
                    return service;
                }));
        }).then(function (service) {
            return Promise.all([
                service,
                self.modelDescriptorService.getUiDescriptorForType(objectType)
            ]).spread(function (service, uiDescriptor) {
                return [
                    {
                        object: service.section,
                        service: service,
                        userInterfaceDescriptor: uiDescriptor,
                        columnIndex: 0,
                        objectType: objectType,
                        path: '/' + encodeURIComponent(sectionDescriptor.id)
                    }
                ];
            }).caught(function (error) {
                console.warn(error.message);
            });
        });
    };
    AbstractRoute.prototype.getOld = function (sectionId) {
        this.eventDispatcherService.dispatch('oldSectionChange', sectionId);
    };
    AbstractRoute.prototype.loadSectionsDescriptors = function () {
        if (!AbstractRoute.sectionsDescriptorsPromise) {
            AbstractRoute.sectionsDescriptorsPromise = Promise.resolve(SystemJS.import('data/sections-descriptors.json'));
        }
        return AbstractRoute.sectionsDescriptorsPromise;
    };
    AbstractRoute.prototype.loadSettings = function (sectionId) {
        var _this = this;
        var self = this, objectType = model_1.Model.SectionSettings, columnIndex = 1, parentContext = this.stack[columnIndex - 1], context = {
            columnIndex: columnIndex,
            objectType: objectType,
            parentContext: parentContext,
            path: parentContext.path + '/settings/_/' + encodeURIComponent(sectionId)
        };
        return Promise.all([
            this.modelDescriptorService.getUiDescriptorForType(objectType)
        ]).spread(function (uiDescriptor) {
            context.object = parentContext.object.settings;
            context.userInterfaceDescriptor = uiDescriptor;
            return self.updateStackWithContext(_this.stack, context);
        });
    };
    AbstractRoute.getPathSection = function (path) {
        return _.head(_.compact(_.split(path, '/')));
    };
    return AbstractRoute;
}());
AbstractRoute.sectionsServices = new Map();
exports.AbstractRoute = AbstractRoute;
function Route(path) {
    return function (target, propertyKey) {
        crossroads.addRoute(path, function () {
            var params = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                params[_i] = arguments[_i];
            }
            return target[propertyKey].apply(target.constructor.getInstance(), params);
        });
    };
}
exports.Route = Route;
