"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var redux_1 = require("redux");
var main_1 = require("../reducers/main");
var middleware_client_1 = require("./middleware-client");
var event_dispatcher_service_1 = require("./event-dispatcher-service");
var _ = require("lodash");
var DatastoreService = (function () {
    function DatastoreService() {
        var self = this, composeEnhancers = window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ || redux_1.compose;
        this.store = redux_1.createStore(main_1.dispatchAction, composeEnhancers(redux_1.applyMiddleware(this.promiseMiddleware)));
        this.store.subscribe(function () { return self.handleModelChange(); });
        this.eventDispatcherService = event_dispatcher_service_1.EventDispatcherService.getInstance();
        this.eventDispatcherService.addEventListener('middlewareModelChange', function (args) { return self.handleMiddlewareModelChange(args); });
        this.middlewareClient = middleware_client_1.MiddlewareClient.getInstance();
    }
    DatastoreService.getInstance = function () {
        if (!DatastoreService.instance) {
            DatastoreService.instance = new DatastoreService();
        }
        return DatastoreService.instance;
    };
    DatastoreService.prototype.query = function (type, methodName, idPath, middlewareCriteria) {
        var _this = this;
        return this.middlewareClient.callRpcMethod(methodName, middlewareCriteria).then(function (result) {
            var payload = Array.isArray(result) ? result : [result];
            _this.store.dispatch({
                type: main_1.ACTIONS.IMPORT_OBJECTS,
                meta: {
                    type: type,
                    idPath: idPath
                },
                payload: payload
            });
            return payload;
        });
    };
    DatastoreService.prototype.stream = function (type, methodName, idPath, partial, middlewareCriteria) {
        var _this = this;
        // TODO: count rpc call in order to have the total of object for a type (LIMIT: 2000)
        return this.middlewareClient.callRpcMethod(methodName, middlewareCriteria).then(function (message) {
            var streamId = message.id, stream = _this.getDefaultStreamObject(type, streamId, idPath, partial);
            _this.store.dispatch({
                type: main_1.ACTIONS.SAVE_STREAM,
                meta: {
                    type: streamId
                },
                payload: stream
            });
            return _this.handleFragmentResponse(_this.getState().get('streams').get(streamId), message);
        });
    };
    DatastoreService.prototype.getDefaultStreamObject = function (type, streamId, idPath, partial) {
        return {
            type: type,
            streamId: streamId,
            idPath: idPath,
            startSequence: 1,
            endSequence: 1,
            lastSequence: 1,
            partial: partial,
            reachEnd: false,
            data: []
        };
    };
    DatastoreService.prototype.handleFragmentResponse = function (stream, message) {
        var fragment = message.args.fragment || [], reachEnd = message.name === 'end', sequenceNumber = message.args.seqno, type = stream.get('type'), idPath = stream.get('idPath');
        if (reachEnd) {
            stream = stream.set('lastSequence', sequenceNumber)
                .set('reachEnd', true);
            this.store.dispatch({
                type: main_1.ACTIONS.SAVE_STREAM,
                meta: {
                    type: message.id
                },
                payload: stream.toJS()
            });
            return stream;
        }
        var payload = _.castArray(fragment);
        // TODO: Store only data when the total number is under or equal to 2000.
        this.store.dispatch({
            type: main_1.ACTIONS.IMPORT_OBJECTS,
            meta: {
                type: type,
                idPath: idPath
            },
            payload: payload
        });
        var dataStore = this.getState().get(type), previousLastSequence = stream.get('lastSequence'), data = payload.map(function (value) {
            return dataStore.get(_.get(value, idPath)).toJS();
        });
        if (sequenceNumber > previousLastSequence) {
            stream = stream.set('lastSequence', sequenceNumber);
        }
        // TODO: Remove data from strea objects. (related to montage data)
        stream = stream.set('endSequence', sequenceNumber)
            .set('reachEnd', false)
            .set('data', stream.get('data').concat(data));
        this.store.dispatch({
            type: main_1.ACTIONS.SAVE_STREAM,
            meta: {
                type: message.id
            },
            payload: stream.toJS()
        });
        if (stream.get('partial')) {
            return this.getState().get('streams').get(stream.get('streamId'));
        }
        return this.getNextSequenceForStream(stream.get('streamId'));
    };
    DatastoreService.prototype.getNextSequenceForStream = function (streamId) {
        return this.getSequenceForStream(streamId);
    };
    DatastoreService.prototype.getPreviousSequenceForStream = function (streamId) {
        return this.getSequenceForStream(streamId, false);
    };
    DatastoreService.prototype.getSequenceForStream = function (streamId, next) {
        var _this = this;
        if (next === void 0) { next = true; }
        next = !!next;
        if (streamId) {
            var stream_1 = this.getState().get('streams').get(streamId);
            if (stream_1) {
                if (stream_1.get('reachEnd')) {
                    return Promise.resolve(stream_1);
                }
                var currentEndSequence = stream_1.get('endSequence'), sequenceNumber = next ? currentEndSequence + 1 : currentEndSequence - 1;
                if (sequenceNumber > 0) {
                    return this.middlewareClient.continueRpcMethod(streamId, sequenceNumber).then(function (message) {
                        return _this.handleFragmentResponse(stream_1, message);
                    });
                }
                return Promise.reject('The sequence number must equal or greater than 1');
            }
            return Promise.reject('Stream can be found with stream ID: ' + streamId);
        }
        return Promise.reject('Stream ID missing');
    };
    DatastoreService.prototype.save = function (type, id, object) {
        this.store.dispatch({
            type: main_1.ACTIONS.SAVE_OBJECT,
            meta: {
                type: type,
                id: id
            },
            payload: object
        });
    };
    DatastoreService.prototype.delete = function (type, id) {
        var state = this.store.getState();
        if (state && state.get(type) && state.get(type).has(id)) {
            this.store.dispatch({
                type: main_1.ACTIONS.DELETE_OBJECT,
                meta: {
                    type: type,
                    id: id
                }
            });
        }
    };
    DatastoreService.prototype.import = function (type, objects) {
        this.store.dispatch({
            type: main_1.ACTIONS.IMPORT_OBJECTS,
            meta: {
                type: type
            },
            payload: objects
        });
    };
    DatastoreService.prototype.rename = function (type, oldId, newId) {
        this.store.dispatch({
            type: main_1.ACTIONS.RENAME_OBJECT,
            meta: {
                type: type,
                id: oldId
            },
            payload: newId
        });
    };
    DatastoreService.prototype.saveOverlay = function (type, id, taskId, payload) {
        this.store.dispatch({
            type: main_1.ACTIONS.SAVE_OVERLAY,
            meta: {
                type: type,
                id: id,
                taskId: taskId
            },
            payload: payload
        });
    };
    DatastoreService.prototype.deleteOverlay = function (type, id, taskId) {
        this.store.dispatch({
            type: main_1.ACTIONS.DELETE_OVERLAY,
            meta: {
                type: type,
                id: id,
                taskId: taskId
            }
        });
    };
    DatastoreService.prototype.getState = function () {
        return this.store.getState();
    };
    DatastoreService.prototype.handleMiddlewareModelChange = function (args) {
        var type = _.upperFirst(_.camelCase(args.service)), ids = args.ids, entities = args.entities || [args.data], operation = args.operation;
        switch (operation) {
            case 'create':
            case 'update':
                for (var _i = 0, entities_1 = entities; _i < entities_1.length; _i++) {
                    var entity = entities_1[_i];
                    this.save(type, entity.id, entity);
                }
                break;
            case 'delete':
                for (var _a = 0, ids_1 = ids; _a < ids_1.length; _a++) {
                    var id = ids_1[_a];
                    this.delete(type, id);
                }
                break;
            case 'rename':
                for (var _b = 0, ids_2 = ids; _b < ids_2.length; _b++) {
                    var id = ids_2[_b];
                    this.rename(type, id[0], id[1]);
                }
                break;
            default:
                console.log('Unknown operation: "' + operation + '"');
                break;
        }
    };
    DatastoreService.prototype.handleModelChange = function () {
        this.eventDispatcherService.dispatch('stateChange', this.store.getState());
    };
    // DTM (should be replaced by redux-promise as soon as mr is out)
    DatastoreService.prototype.promiseMiddleware = function (_ref) {
        var dispatch = _ref.dispatch;
        return function (next) {
            return function (action) {
                return Promise.is(action.payload) ?
                    action.payload.then(function (result) { return dispatch(Object.assign({}, action, { payload: result })); }, function (error) { return dispatch(Object.assign({}, action, { payload: error, error: true })); }) :
                    next(action);
            };
        };
    };
    return DatastoreService;
}());
exports.DatastoreService = DatastoreService;
